#ifndef UML_DATATYPES_H
#define UML_DATATYPES_H

#include <string>
#include <list>
#include <set>
#include <utl.h>

typedef bool UMLBoolean;
typedef long UMLInteger;

/**
 * A gentle hack of the standard to include an
 * "unlimited value" in order to avoid the definition
 * of an unlimited integer type. Since unlimited integers
 * are really only applied to multiplicity, we can safely
 * choose -1 as a representative value.
 */
static const UMLInteger mr_unlimited = -1;

typedef std::string UMLName;
typedef std::set<UMLName> UMLNameBag;

typedef std::string UMLString;
typedef std::set<UMLString> UMLStringBag;


/**
 * UML 1.4 section 2.7.2.29
 */
enum UMLVisibilityKind {
  vk_public,		//@< Other elements may use the target
  vk_protected,		//@< Descendant elements may use the target
  vk_private,		//@< Only the source element may use the target
  vk_package		//@< Only elements in the same package may use the target
};

/**
 * UML 1.4 section 2.7.2.23
 */
enum UMLScopeKind {
  sk_instance,		//@< Feature pertains to instances of a classifier
  sk_classifier		//@< Feature pertains to an entire classifier
};

/**
 * UML 1.4 section 2.7.2.7
 */
enum UMLChangeabilityKind {
  ck_changeable,	//@< The value may be changed
  ck_frozen,		//@< The value may not be changed from the source end
  ck_addOnly		//@< Values may be added if multiplicity not fixed
};

/**
 * UML 1.4 section 2.7.2.19
 */
enum UMLOrderingKind {
  ok_unordered,
  ok_ordered
};

/**
 * UML 1.4 section 2.7.2.2
 */
enum UMLAggregationKind {
  ak_none,		//@< No aggregation
  ak_aggregate,		//@< Shared ownership
  ak_composite		//@< Total ownership
};

/**
 * UML 1.4 section 2.7.2.20
 */
enum UMLParameterDirectionKind {
  pdk_in,		//@< Parameter is only an input
  pdk_inout,		//@< Parameter is both an input and output
  pdk_out,		//@< Parameter is only an output
  pdk_return		//@< Parameter is a return value
};

/**
 * UML 1.4 section 2.7.2.6
 */
enum UMLCallConcurrencyKind {
  cck_sequential,	//@< Only one call to an instance at a time
  cck_guarded,		//@< Concurrent calls, only one active
  cck_concurrent	//@< Concurrent calls, all active
};

/**
 * UML 1.4 section 2.7.2.16
 *
 * @todo Make this pass by value
 */
class UMLMultiplicityRange
{
 public:
  UMLMultiplicityRange();
  ~UMLMultiplicityRange();

  utl::single_attribute<UMLInteger> lower;
  utl::single_attribute<UMLInteger> upper;
};

/**
 * UML 1.4.section 2.7.2.15
 *
 * @todo Make this pass by value
 */
class UMLMultiplicity
{
 public:
  UMLMultiplicity();
  ~UMLMultiplicity();

  utl::unordered_attribute<UMLMultiplicityRange> range;
};

/**
 * UML 1.4 section 2.7.2.8
 *
 * @todo Make this pass by value
 */
class UMLExpression
{
 public:
  UMLExpression();
  virtual ~UMLExpression();

  utl::single_attribute<UMLName> language;
  utl::single_attribute<UMLString> body;
};

/*
 * Because derived expressions aren't interesting (yet), just
 * typedef them. I doubt these will ever be more interesting.
 */
typedef UMLExpression UMLActionExpression;
typedef UMLExpression UMLBooleanExpression;
typedef UMLExpression UMLMappingExpression;
typedef UMLExpression UMLProcedureExpression;
typedef UMLExpression UMLTypeExpression;
typedef UMLExpression UMLArgListExpression;
typedef UMLExpression UMLIterationExpression;
typedef UMLExpression UMLObjectSetExpression;
typedef UMLExpression UMLTimeExpression;

#endif
