#ifndef UTL_ATTRIBUTE_H
#define UTL_ATTRIBUTE_H

#include <list>
#include <set>
#include <utl_exception.h>

namespace utl
{
  /**
   * Attribute base class
   */
  class attribute
  {
  public:
    attribute()
    {}

    virtual ~attribute()
    {}
  };

  /**
   * Because I'm feeling very, very, very lazy. I've given up
   * on the idea of correct implementation for attributes.
   * From now on, you can unset them all. It reduces code and
   * confusion.
   */
  template<class Type>
  class single_attribute : public attribute
  {
  public:
    typedef Type type;

    single_attribute() :
      attribute()
    {}

    single_attribute(const type &val) :
      attribute(),
      _attr(val)
    {}

    virtual ~single_attribute()
    {}

    const type &get() const
    {
      if(!_set) throw not_set();
      return _attr;
    }

    void set(const type &val)
    {
      _attr = val;
      _set = true;
    }

    void unset()
    {
      _set = false;
    }

  private:
    type _attr;
    bool _set;
  };


  /**
   * Ordered, multiple attributes.
   */
  template<class Type>
  class ordered_attribute : public attribute
  {
  public:
    typedef Type type;
    typedef std::list<Type> container;

    ordered_attribute() :
      attribute()
    {}

    ordered_attribute(const container &cont) :
      attribute(),
      _container(cont)
    {}

    virtual ~ordered_attribute()
    {
      _container.clear();
    }

    const container &get() const
    {
      if(!_set) throw not_set();
      return _container;
    }

    void set(const container &val)
    {
      _container = val;
      _set = true;
    }

    void add(const type &val)
    {
      _container.push_back(val);
      _set = true;
    }

    void add_before(const type &new_val,
		    const type &before_val)
    {
      typename container::iterator i = _container.find(before_val);
      if(i != _container.end()) {
	_container.insert(i, new_val);
      }
      _set = true;
    }

    void remove(const type &val)
    {
      if(!_set) throw not_set();
      _container.remove(val);
    }

    void modify(const type &old_val,
		const type &new_val)
    {
      if(!_set) throw not_set();
      typename container::iterator i = _container.find(old_val);
      if(i != _container.end()) {
	*i = new_val;
      }
    }

    void unset()
    {
      _container.clear();
      _set = false;
    }

  private:
    container _container;
    bool _set;
  };


  /**
   * Unordered, multiple attributes.
   */
  template<class Type>
  class unordered_attribute : public attribute
  {
  public:
    typedef Type type;
    typedef std::set<Type> container;

    unordered_attribute() :
      attribute()
    {}

    unordered_attribute(const container &cont) :
      attribute(),
      _container(cont)
    {}

    virtual ~unordered_attribute()
    {
      _container.clear();
    }

    const container &get() const
    {
      if(!_set) throw not_set();
      return _container;
    }

    void set(const container &val)
    {
      _container = val;
      _set = true;
    }

    void add(const type &val)
    {
      _container.push_back(val);
      _set = true;
    }

    void remove(const type &val)
    {
      if(!_set) throw not_set();
      _container.remove(val);
    }

    void modify(const type &old_val,
		const type &new_val)
    {
      if(!_set) throw not_set();
      typename container::iterator i = _container.find(old_val);
      if(i != _container.end()) {
	*i = new_val;
      }
    }

    void unset()
    {
      _container.clear();
      _set = false;
    }

  private:
    container _container;
    bool _set;
  };

};

#endif
