#ifndef UTL_RELATION_H
#define UTL_RELATION_H

#include <utl_exception.h>
#include <utl_connection.h>

namespace utl
{
  /**
   * Relation base class
   */
  class relation
  {
  public:
    relation()
    {}

    virtual ~relation()
    {}
  };

  /**
   * Because I'm feeling very, very, very lazy. I've given up
   * on the idea of correct implementation for relations.
   * From now on, you can unset them all. It reduces code and
   * confusion.
   *
   * The set method takes something a little different than
   * attributes. It takes an endpoint. An endpoint identifies
   * the other element and its corresponding attribute that
   * define the relationship.
   */
  template<class Type>
  class single_relation : public relation
  {
  public:
    typedef Type type;

    single_relation() :
      relation()
    {}

    single_relation(const type &val) :
      relation(),
      _rel(val)
    {}

    virtual ~single_relation()
    {}

    /**
     * Just return the elment in questions - not
     * the entire endpoint.
     */
    const type &get() const
    {
      if(!_set) throw not_set();
      return _rel;
    }

    /**
     * The set operation only operates on endpoints.
     */
    void set(const type &val)
    {
      _rel = val;
      _set = true;
    }

    void unset()
    {
      _set = false;
    }

    /**
     * The connect method just sets the value.
     */
    void connect(endpoint<Type> &ep)
    {
      set(ep._elem);
    }

    /**
     * The disconnect method just unsets the
     * value.
     */
    void disconnect(endpoint<Type> &ep)
    {
      unset();
    }

  private:
    type _rel;
    bool _set;
  };


  /**
   * Ordered, multiple relations.
   */
  template<class Type>
  class ordered_relation : public relation
  {
  public:
    typedef Type type;
    typedef std::list<Type> container;

    ordered_relation() :
      relation()
    {}

    ordered_relation(const container &cont) :
      relation(),
      _container(cont)
    {}

    virtual ~ordered_relation()
    {
      _container.clear();
    }

    const container &get() const
    {
      if(!_set) throw not_set();
      return _container;
    }

    void set(const container &val)
    {
      _container = val;
      _set = true;
    }

    void add(const type &val)
    {
      _container.push_back(val);
      _set = true;
    }

    void add_before(const type &new_val,
		    const type &before_val)
    {
      typename container::iterator i = _container.find(before_val);
      if(i != _container.end()) {
	_container.insert(i, new_val);
      }
      _set = true;
    }

    void remove(const type &val)
    {
      if(!_set) throw not_set();
      _container.remove(val);
    }

    void modify(const type &old_val,
		const type &new_val)
    {
      if(!_set) throw not_set();
      typename container::iterator i = _container.find(old_val);
      if(i != _container.end()) {
	*i = new_val;
      }
    }

    void unset()
    {
      _container.clear();
      _set = false;
    }

    /**
     * The connect method just adds the value.
     */
    void connect(endpoint<Type> &ep)
    {
      add(ep._elem);
    }

    /**
     * The disconnect method just removes the
     * value.
     */
    void disconnect(endpoint<Type> &ep)
    {
      remove(ep._elem);
    }

  private:
    container _container;
    bool _set;
  };


  /**
   * Unordered, multiple relations.
   */
  template<class Type>
  class unordered_relation : public relation
  {
  public:
    typedef Type type;
    typedef std::set<Type> container;

    unordered_relation() :
      relation()
    {}

    unordered_relation(const container &cont) :
      relation(),
      _container(cont)
    {}

    virtual ~unordered_relation()
    {
      _container.clear();
    }

    const container &get() const
    {
      if(!_set) throw not_set();
      return _container;
    }

    void set(const container &val)
    {
      _container = val;
      _set = true;
    }

    void add(const type &val)
    {
      _container.push_back(val);
      _set = true;
    }

    void remove(const type &val)
    {
      if(!_set) throw not_set();
      _container.remove(val);
    }

    void modify(const type &old_val,
		const type &new_val)
    {
      if(!_set) throw not_set();
      typename container::iterator i = _container.find(old_val);
      if(i != _container.end()) {
	*i = new_val;
      }
    }

    void unset()
    {
      _container.clear();
      _set = false;
    }

    /**
     * The connect method just adds the value.
     */
    void connect(endpoint<Type> &ep)
    {
      add(ep._elem);
    }

    /**
     * The disconnect method just removes the
     * value.
     */
    void disconnect(endpoint<Type> &ep)
    {
      remove(ep._elem);
    }

  private:
    container _container;
    bool _set;
  };

};


#endif
